/*
 * This program counts the number of prime numbers up to first the number on the command line, using a number of threads given as the second number on the command line, using a slightly better primality test
 */
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>

int n = 1000;
int nthreads = 24;

#define MAX_THREADS	64

#define LIST_SIZE	1000
int list_size;

// a list of prime numbers generated when the program starts

int primes_list[LIST_SIZE];

int is_prime (int p) {
	if (p == 0) return false;
	if (p == 1) return false;
	int lim = sqrt (p);

	// if the number is divisible by one of the known primes we stored, it is not prime

	int i;
	for (i=0; i < list_size && primes_list[i] <= lim; i++) {
		if (p % primes_list[i] == 0) return false;
	}

	// maybe we have to check numbers we didn't store, but most numbers don't arrive here

	for (i=list_size ? primes_list[list_size-1] : 2; i<=lim; i++) {
		if (p % i == 0) return false;
	}
	return true;
}

struct counter {
	int x;
};

counter counters[MAX_THREADS];

void *partial_pi (void *p) {
	counter *my_counter = (counter *) p;
	int *my_number = &my_counter->x;
	int a = *my_number * (n/nthreads);
	int b = ((*my_number)+1) * (n/nthreads);
	if (*my_number == nthreads-1) b = n;
	*my_number = 0;
	for (int i=a; i<b; i++) (*my_number) += is_prime (i);
        return NULL;
}

void fill_list (void) {
	int p = 2;
	list_size = 0;
	do {
		if (is_prime (p)) primes_list[list_size++] = p;
		p++;
	} while (list_size < LIST_SIZE);
}

int main (int argc, char **argv) {
	if (argc == 3) {
		sscanf (argv[1], "%d", &n);
		sscanf (argv[2], "%d", &nthreads);
		assert (nthreads <= MAX_THREADS);
	}
	fill_list ();
	pthread_t threads[nthreads];

	for (int z=0; z<nthreads; z++) {
		counters[z].x = z;
		pthread_create(&threads[z], NULL, (void *(*)(void*)) partial_pi, (void*) &counters[z]);
	}
	for (int z=0; z<nthreads; z++) {
		pthread_join(threads[z], NULL);
	}
	int sum = 0;
	for (int i=0; i<nthreads; i++) sum += counters[i].x;
	printf ("%d\n", sum);
	return 0;
}
